/* Faster but has to be undefined when debugging. */
#define G_DISABLE_CAST_CHECKS

#include <gtk/gtk.h>

#ifdef HAVE_CONFIG_H
#   include "../config.h"
#endif

#ifdef ENABLE_NLS
#   include <libintl.h>
#   define _(str) (gettext((str)))
#else
#   define _(str) (str)
#endif

/* Just to mark constant translatable strings, like in rcfile.c. */
#define N_(str) (str)

#if HAVE_ISO_C_VARARGS

#   define ERROR_MSG(...)                       \
    do {                                        \
        gchar *message;                         \
                                                \
        message = g_strdup_printf(__VA_ARGS__); \
        g_printerr(message);                    \
        g_free(message);                        \
    } while (0)

#   define g_build_filename(...) \
    clean_filename_free(g_build_filename(__VA_ARGS__))

#elif HAVE_GNU_C_VARARGS

#   define ERROR_MSG(format...)              \
    do {                                     \
        gchar *message;                      \
                                             \
        message = g_strdup_printf(format);   \
        g_printerr(message);                 \
        g_free(message);                     \
    } while (0)

#   define g_build_filename(str...) \
    clean_filename_free(g_build_filename(str))

#else
#   error "No variadic macros"
#endif

/* Instead of finding it in a system or glib header. */
#define PI 3.1415926535897932384626433832795029L

/* User configurable settings. */

#define ZOOM_FACTOR    1.1
#define MOVE_OFFSET    20.0
#define BIG_ROTATION   (PI / 4.0)       /* 45 deg */
#define SMALL_ROTATION (PI / 1800.0)    /* 0.1 deg */
#define MIPMAP_RATIO   0.75

/* Font for the help box. */
#define FONT "courier Medium 12"


/* End of user configurable settings. */

/* Loader according to filename extension. */
typedef enum {
    LOADER_NONE,                /* Unrecognized extension. */
    LOADER_NORMAL,              /* GdkPixbuf, then ImageMagick. */
    LOADER_DECOMP               /* ImageMagick with decompression. */
} loader_t;


struct format {
    const gchar *name;
    const loader_t loader;
};

/* Bits for the refresh() function. */
#define REFRESH_IMAGE      (1 << 0)
#define REFRESH_NOW        (1 << 1)
#define REFRESH_STATUS     (1 << 2)
#define APPEND_HISTORY     (1 << 3)

/* Infos that can be freed once the map is created. */
typedef struct {
    gint width;
    gint height;
    gint x_tiles;
    gint y_tiles;
    GdkPixbuf *pixbuf;
} loading_data;

typedef struct {
    gfloat x0;
    gfloat y0;
    gfloat x1;
    gfloat y1;
} tile_dim;

typedef struct {
    gint nb_tiles;
    tile_dim *tiles;
    guint *tex_ids;
    guint list;
    loading_data *map_init;
} texture_map;

/* The gliv_image structure. */
typedef struct {
    GList *node;
    gint number;
    gchar *ident;
    gint width;
    gint height;
    gint nb_maps;
    texture_map *maps;
    gboolean has_alpha;
} gliv_image;

/* Options controlled by the user. */
typedef struct {
    gboolean fullscreen;
    gboolean maximize;
    gboolean scaledown;
    gboolean menu_bar;
    gboolean status_bar;
    gboolean scrollbars;
    gboolean zoom_pointer;
    gboolean alpha_checks;
    gboolean dither;
    gboolean force;
    gboolean build_menus;
    gboolean mipmap;
    gboolean mnemonics;
    gboolean loop;
    gint delay;
    gint duration;
    gint history_size;
    gint bg_col[3];
    gint alpha1[3];
    gint alpha2[3];
} options_struct;

/* Some global flags and variables. */
typedef struct {
    gboolean cursor_hidden;
    gboolean help;
    gboolean alpha_checks_changed;
    gboolean filtering_enabled;
    gint max_texture_size;
    GtkAllocation *wid_size;
    gint scr_width;
    gint scr_height;
} rt_struct;

/* callbacks.c */
void install_callbacks(gpointer object);
void process_events(void);

/* files_list.c */
gint get_list_length(void);
GList *get_list_head(void);
GList *get_list_end(void);
void remove_from_list(GList * node);
void reorder_files(gboolean shuffle);
gchar **get_sorted_files_array(void);
void init_list(gchar ** names, gint num, gboolean recursive,
               gboolean shuffle, gboolean sort);
void place_next(const gchar * filename);
gboolean insert_after_current(gchar ** names, gint nb, gboolean shuffle);
void confirm_remove_current(void);
void delete_selected_image(void);

/* foreach_file.c */
typedef void (*foreach_file_func) (const gchar * path);
void foreach_file(const gchar * path, foreach_file_func func);

/* formats.c */
const struct format *ext_to_loader(const gchar * str, guint len);

/* gl_widget.c */
void update_bg_color(void);
gboolean toggle_cursor(gboolean visible);
void set_hide_cursor_enabled(gboolean enabled);
void schedule_hide_cursor(void);
void create_gl_widget(void);

/* gliv.c */
G_GNUC_NORETURN void quit(gint code);

/* help.c */
gboolean toggle_help(void);
void show_about_box(void);

/* history.c */
void init_history(void);
void append_history(void);
void undo(void);
void redo(void);
void clear_history(void);

/* imagemagick.c */
GdkPixbuf *load_with_imagemagick(const gchar * filename);
gboolean have_imagemagick(void);
void imagemagick_finish(void);

/* images_menus.c */
void set_rebuilding_entry(GtkMenuItem * item);
void rebuild_directories(GtkMenuItem * root);
void rebuild_images(GtkMenuItem * root);
void rebuild_images_menus(void);

/* loading.c */
gboolean is_loading(void);
void destroy_image(gliv_image * im);
void install_segv_handler(void);
void fill_ident(gliv_image * im);
gboolean is_loadable(const gchar * filename);
GdkPixbuf *load_gdk_pixbuf(const gchar * filename);
gliv_image *load_file(const gchar * filename);

/* matrix.c */
G_GNUC_PURE gfloat get_matrix_zoom(void);
G_GNUC_PURE gfloat get_matrix_angle(void);
G_GNUC_PURE gboolean float_equal(gfloat a, gfloat b);
void get_matrix_bounding_box(gfloat * min_x, gfloat * max_x,
                             gfloat * min_y, gfloat * max_y);
void write_gl_matrix(void);
void matrix_cpy(gfloat * dest, gfloat * src);
gboolean matrix_tile_visible(tile_dim * tile);
G_GNUC_PURE gboolean is_matrix_symmetry(void);
G_GNUC_PURE gboolean get_matrix_has_changed(void);
G_GNUC_PURE gboolean is_filtering_needed(void);
gboolean matrix_set_max_zoom(gint width, gint height, gboolean do_it);
void matrix_reset(void);
void matrix_rotate(gfloat angle);
void matrix_move(gfloat x, gfloat y);
void matrix_zoom(gfloat ratio, gfloat x, gfloat y);
void matrix_flip_h(void);
void matrix_flip_v(void);

/* menus.c */
void set_menu_label(GtkMenuItem * item, const gchar * text, gboolean mnemonic);
GtkAccelGroup *create_menus(void);

/* move_pointer.c */
void move_pointer(gint dest_x, gint dest_y);

/* next_image.c */
void after_reorder(gint new_id);
gboolean load_direction(gint dir);
void open_next_image(void);
void load_first_image(void);
void load_second_image(void);
void open_file(const gchar * filename, gboolean add_dir, gboolean shuffle_flag);
void reload_all_images(void);
void menu_load(const gchar * filename);
void unload(GList * node);
void set_slide_show_menu(GtkMenuItem * item);
void start_slide_show(void);
gboolean slide_show_started(void);
void toggle_slide_show(void);
void begin_with_slide_show(void);

/* open_dialog.c */
gboolean menu_open(void);

/* options.c */
void show_options(void);

/* rcfile.c */
options_struct *load_rc(gboolean from_file);
void load_accelerators(void);
void save_rc(options_struct * opt);

/* rendering.c */
void render(void);
void zoom_in(gfloat ratio);
void refresh(gint what);

/* str_utils.c */
const gchar *add_mnemonic(const gchar * str);
const gchar *locale_to_utf8(const gchar * str);
gchar *clean_filename(const gchar * filename);
gchar *clean_filename_free(gchar * filename);
gchar *fast_build_filename(const gchar * dir, const gchar * file);

/* textures.c */
void prioritize_textures(gliv_image * im, gboolean is_prio);
void create_maps(gliv_image * im);

/* windows.c */
void show_dialog(GtkWidget * widget, const gchar * name);
void set_loading_entry(const gchar * filename);
void toggle_menu_bar(void);
void toggle_status_bar(void);
void toggle_scrollbars(void);
void update_status_bar(void);
void update_scrollbars(gboolean redraw);
void refresh_all(void);
void create_windows(void);
void goto_window(gboolean force_resize);
void toggle_fullscreen(gboolean enable);
void update_window_title(void);
GtkWindow *get_current_window(void);

/* zoom_frame.c */
void draw_zoom_frame(gboolean draw);
void set_zoom_frame(gint x, gint y, gint width, gint height);
void clear_zoom_frame(void);
void zoom_frame(void);
